@@ -1,6 +1,6 @@
-# -*- coding: utf-8 -*-
+#!/usr/bin/env python3
 #
-# Copyright (C) 2009-2014 Sébastien Helleu <flashcode@flashtux.org>
+# Copyright (C) 2009-2021 Sébastien Helleu <flashcode@flashtux.org>
 #
 # This file is part of msgcheck.
 #
@@ -15,12 +15,10 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with msgcheck.  If not, see <http://www.gnu.org/licenses/>.
+# along with msgcheck.  If not, see <https://www.gnu.org/licenses/>.
 #
 
-"""
-Perform various checks on gettext files.
-"""
+"""Gettext file checker."""
 
-from . import msgcheck
-from . import po
+from . import msgcheck  # noqa: F401
+from . import po  # noqa: F401
