@@ -1,6 +1,6 @@
-# -*- coding: utf-8 -*-
+#!/usr/bin/env python3
 #
-# Copyright (C) 2009-2014 Sébastien Helleu <flashcode@flashtux.org>
+# Copyright (C) 2009-2021 Sébastien Helleu <flashcode@flashtux.org>
 #
 # This file is part of msgcheck.
 #
@@ -15,44 +15,43 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with msgcheck.  If not, see <http://www.gnu.org/licenses/>.
+# along with msgcheck.  If not, see <https://www.gnu.org/licenses/>.
 #
 
-"""
-Classes to read and check PO (gettext) files.
-"""
+"""Classes to read and check PO (gettext) files."""
 
 from __future__ import print_function
 
 from codecs import escape_decode
 import os
 import re
-import subprocess
-import sys
+import subprocess  # nosec
+import tempfile
 
 # enchant module is optional, spelling is checked on demand
 # (argument -s/--spell)
-ENCHANT_FOUND = False
 try:
     from enchant import Dict, DictWithPWL, DictNotFoundError
     from enchant.checker import SpellChecker
     ENCHANT_FOUND = True
 except ImportError:
-    pass
+    ENCHANT_FOUND = False
+
+from . utils import count_lines, replace_formatters
 
 
-def count_lines(string):
-    """
-    Count the number of lines in a string or translation.
-    """
-    count = len(string.split('\n'))
-    if count > 1 and string.endswith('\n'):
-        count -= 1
-    return count
+def get_concatenated_files(filenames):
+    """Return concatenated content of multiple files."""
+    if not filenames:
+        return None
+    content = []
+    for filename in filenames:
+        with open(filename, 'rb') as _file:
+            content.append(_file.read().decode('utf-8'))
+    return '\n'.join(content)
 
 
-# pylint: disable=too-few-public-methods
-class PoReport(object):
+class PoReport:  # pylint: disable=too-few-public-methods
     """A message in report (commonly an error in detected in gettext file)."""
 
     # pylint: disable=too-many-arguments
@@ -70,17 +69,13 @@ class PoReport(object):
         if self.idmsg == 'extract':
             return self.message + '\n---'
         if self.idmsg == 'compile':
-            return '{0}\n{1}'.format('=' * 70, self.message)
-        is_list = type(self.message) is list
-        count = '(%d)' % len(self.message) if is_list else ''
-        msg = '{0}\n{1}:{2}: [{3}{4}] {5}{6}'.format(
-            '=' * 70,
-            self.filename,
-            self.line,
-            self.idmsg,
-            count,
-            '(fuzzy) ' if self.fuzzy else '',
-            ', '.join(self.message) if is_list else self.message)
+            return f'{"=" * 70}\n{self.message}'
+        is_list = isinstance(self.message, list)
+        count = f'({len(self.message)})' if is_list else ''
+        str_fuzzy = '(fuzzy) ' if self.fuzzy else ''
+        str_msg = ', '.join(self.message) if is_list else self.message
+        msg = (f'{"=" * 70}\n{self.filename}:{self.line}: '
+               f'[{self.idmsg}{count}] {str_fuzzy}{str_msg}')
         if self.mid:
             msg += '\n---\n' + self.mid
         if self.mstr:
@@ -89,10 +84,10 @@ class PoReport(object):
 
     def get_misspelled_words(self):
         """Return list of misspelled words."""
-        return self.message if type(self.message) is list else []
+        return self.message if isinstance(self.message, list) else []
 
 
-class PoMessage(object):
+class PoMessage:
     """
     A message from a gettext file. It is stored as a list of tuples
     (string, translation).
@@ -119,24 +114,21 @@ class PoMessage(object):
     """
 
     # pylint: disable=too-many-arguments
-    def __init__(self, filename, line, msg, charset, fuzzy):
+    def __init__(self, filename, line, msg, charset, fuzzy, fmt, noqa):
         """Build a PO message."""
         self.filename = filename
         self.line = line
         # unescape strings
-        if sys.version_info < (3,):
-            # python 2.x
-            msg = {k: escape_decode(v)[0] for k, v in msg.items()}
-        else:
-            # python 3.x
-            msg = {k: escape_decode(v)[0]. decode(charset)
-                   for k, v in msg.items()}
+        msg = {
+            k: escape_decode(v)[0].decode(charset)
+            for k, v in msg.items()
+        }
         # build messages as a list of tuples: (string, translation)
         self.messages = []
         if 'msgid_plural' in msg:
             i = 0
             while True:
-                key = 'msgstr[{0}]'.format(i)
+                key = f'msgstr[{i}]'
                 if key not in msg:
                     break
                 self.messages.append((msg['msgid_plural'], msg[key]))
@@ -144,6 +136,8 @@ class PoMessage(object):
         else:
             self.messages.append((msg.get('msgid', ''), msg.get('msgstr', '')))
         self.fuzzy = fuzzy
+        self.fmt = fmt
+        self.noqa = noqa
 
     def check_lines(self):
         """
@@ -158,9 +152,16 @@ class PoMessage(object):
             nb_str = count_lines(mstr)
             if nb_id != nb_str:
                 errors.append(
-                    PoReport('number of lines: {0} in string, '
-                             '{1} in translation'.format(nb_id, nb_str),
-                             'lines', self.filename, self.line, mid, mstr))
+                    PoReport(
+                        f'number of lines: {nb_id} in string, '
+                        f'{nb_str} in translation',
+                        'lines',
+                        self.filename,
+                        self.line,
+                        mid,
+                        mstr,
+                    )
+                )
         return errors
 
     def check_punct(self, language):
@@ -172,9 +173,14 @@ class PoMessage(object):
         for mid, mstr in self.messages:
             if not mid or not mstr:
                 continue
-            puncts = [(':', ':'), (';', ';'), (',', ','), ('...', '...')]
+            puncts = [
+                (':', ':'),
+                (';', ';'),
+                (',', ','),
+                ('...', '...'),
+            ]
             # special symbols in some languages
-            if language.startswith('ja'):
+            if language[:2] in ['ja', 'zh']:
                 puncts.append(('.', '。'))
             else:
                 puncts.append(('.', '.'))
@@ -189,17 +195,29 @@ class PoMessage(object):
                     break
                 if match_id and not match_str:
                     errors.append(
-                        PoReport('end punctuation: "{0}" in string, '
-                                 '"{1}" not in translation'
-                                 ''.format(punctid, punctstr),
-                                 'punct', self.filename, self.line, mid, mstr))
+                        PoReport(
+                            f'end punctuation: "{punctid}" in string, '
+                            f'"{punctstr}" not in translation',
+                            'punct',
+                            self.filename,
+                            self.line,
+                            mid,
+                            mstr,
+                        )
+                    )
                     break
                 if not match_id and match_str:
                     errors.append(
-                        PoReport('end punctuation: "{0}" in translation, '
-                                 '"{1}" not in string'
-                                 ''.format(punctstr, punctid),
-                                 'punct', self.filename, self.line, mid, mstr))
+                        PoReport(
+                            f'end punctuation: "{punctstr}" in '
+                            f'translation, "{punctid}" not in string',
+                            'punct',
+                            self.filename,
+                            self.line,
+                            mid,
+                            mstr,
+                        )
+                    )
                     break
         return errors
 
@@ -218,20 +236,31 @@ class PoMessage(object):
                 startout = len(mstr) - len(mstr.lstrip(' '))
                 if startin != startout:
                     errors.append(
-                        PoReport('whitespace at beginning: {0} in string, '
-                                 '{1} in translation'
-                                 ''.format(startin, startout),
-                                 'whitespace', self.filename, self.line, mid,
-                                 mstr))
+                        PoReport(
+                            f'whitespace at beginning: {startin} in '
+                            f'string, {startout} in translation',
+                            'whitespace',
+                            self.filename,
+                            self.line,
+                            mid,
+                            mstr,
+                        )
+                    )
             # check whitespace at end of string
             endin = len(mid) - len(mid.rstrip(' '))
             endout = len(mstr) - len(mstr.rstrip(' '))
             if endin != endout:
                 errors.append(
-                    PoReport('whitespace at end: {0} in string, '
-                             '{1} in translation'.format(endin, endout),
-                             'whitespace', self.filename, self.line, mid,
-                             mstr))
+                    PoReport(
+                        f'whitespace at end: {endin} in string, '
+                        f'{endout} in translation',
+                        'whitespace',
+                        self.filename,
+                        self.line,
+                        mid,
+                        mstr,
+                    )
+                )
         return errors
 
     def check_whitespace_eol(self):
@@ -247,16 +276,21 @@ class PoMessage(object):
             strlines = mstr.split('\n')
             if len(idlines) < 2 or len(idlines) != len(strlines):
                 continue
-            for i in range(0, len(idlines)):
-                endin = len(idlines[i]) - len(idlines[i].rstrip(' '))
+            for i, idline in enumerate(idlines):
+                endin = len(idline) - len(idline.rstrip(' '))
                 endout = len(strlines[i]) - len(strlines[i].rstrip(' '))
                 if (endin > 0 or endout > 0) and endin != endout:
                     errors.append(
-                        PoReport('different whitespace at end of a line: {0} '
-                                 'in string, {1} in translation'
-                                 ''.format(endin, endout),
-                                 'whitespace_eol', self.filename, self.line,
-                                 mid, mstr))
+                        PoReport(
+                            f'different whitespace at end of a line: {endin} '
+                            f'in string, {endout} in translation',
+                            'whitespace_eol',
+                            self.filename,
+                            self.line,
+                            mid,
+                            mstr,
+                        )
+                    )
                     break
         return errors
 
@@ -271,7 +305,10 @@ class PoMessage(object):
         for mid, mstr in self.messages:
             if not mid or not mstr:
                 continue
-            checkers[0].set_text(mstr if spelling == 'str' else mid)
+            text = mstr if spelling == 'str' else mid
+            if self.fmt:
+                text = replace_formatters(text, self.fmt)
+            checkers[0].set_text(text)
             misspelled = []
             for err in checkers[0]:
                 misspelled_word = True
@@ -282,12 +319,92 @@ class PoMessage(object):
                 if misspelled_word:
                     misspelled.append(err.word)
             if misspelled:
-                errors.append(PoReport(misspelled, 'spelling-' + spelling,
-                                       self.filename, self.line, mid, mstr))
+                errors.append(
+                    PoReport(
+                        misspelled,
+                        'spelling-' + spelling,
+                        self.filename,
+                        self.line,
+                        mid,
+                        mstr,
+                    )
+                )
         return errors
 
 
-class PoFile(object):
+class Checker:  # pylint: disable=too-many-instance-attributes
+    """Messages checker."""
+
+    def __init__(self):
+        self.numline = 0
+        self.numline_msgid = 0
+        self.fuzzy = False
+        self.msgfuzzy = False
+        self.noqa = False
+        self.msgnoqa = False
+        self.fmt = None
+        self.msgfmt = None
+        self.msg = {}
+        self.msgcurrent = ''
+        self.oldmsgcurrent = ''
+
+    def check_line(self, line):
+        """Check a line of a PO file."""
+        message = None
+        self.numline += 1
+        if not line:
+            return None
+        if line.startswith('#,'):
+            self.fuzzy = 'fuzzy' in line
+            match = re.search(r'([a-z-]+)-format', line, re.IGNORECASE)
+            self.fmt = match.group(1) if match else None
+        if line.startswith('#'):
+            self.noqa = self.noqa or 'noqa' in line
+            return None
+        if line.startswith('msg'):
+            match = re.match(
+                r'([a-zA-Z0-9-_]+(\[\d+\])?)[ \t](.*)',
+                line)
+            if match:
+                self.oldmsgcurrent = self.msgcurrent
+                self.msgcurrent = match.group(1)
+                line = match.group(3)
+                if self.msgcurrent == 'msgid':
+                    if self.oldmsgcurrent.startswith('msgstr'):
+                        message = (
+                            self.numline_msgid,
+                            self.msgfuzzy,
+                            self.msgfmt,
+                            self.msgnoqa,
+                            self.msg,
+                        )
+                    self.msgfuzzy = self.fuzzy
+                    self.msgnoqa = self.noqa
+                    self.fuzzy = False
+                    self.noqa = False
+                    self.msgfmt = self.fmt
+                    self.fmt = None
+                    self.msg = {}
+                    self.numline_msgid = self.numline
+        if self.msgcurrent and line.startswith('"'):
+            self.msg[self.msgcurrent] = (self.msg.get(self.msgcurrent, '') +
+                                         line[1:-1])
+        return message
+
+    def last_check(self):
+        """Consume the last message (after all lines were read)."""
+        if self.msgcurrent.startswith('msgstr'):
+            return (
+                self.numline_msgid,
+                self.msgfuzzy,
+                self.msgfmt,
+                self.msgnoqa,
+                self.msg,
+            )
+        return None
+
+
+class PoFile:
     """
     A gettext file. It includes methods to read the file, and perform
     checks on the translations.
@@ -302,11 +419,12 @@ class PoFile(object):
         }
         self.msgs = []
 
-    def _add_message(self, numline_msgid, msgfuzzy, msg):
+    # pylint: disable=too-many-arguments
+    def _add_message(self, numline_msgid, fuzzy, fmt, noqa, msg):
         """
         Add a message from PO file in list of messages.
         """
-        if 'msgid' in msg and len(msg['msgid']) == 0:
+        if 'msgid' in msg and not msg['msgid']:
             # find file language/charset in properties
             # (first string without msgid)
             match = re.search(r'language: ([a-zA-Z-_]+)',
@@ -319,49 +437,22 @@ class PoFile(object):
             if match:
                 self.props['charset'] = match.group(1)
         self.msgs.append(PoMessage(self.filename, numline_msgid, msg,
-                                   self.props['charset'], msgfuzzy))
+                                   self.props['charset'], fuzzy, fmt, noqa))
 
-    def read(self):
+    def read(self):  # pylint: disable=too-many-locals
         """
         Read messages in PO file.
         """
         self.msgs = []
-        (numline, numline_msgid) = (0, 0)
-        (fuzzy, msgfuzzy) = (False, False)
-        msg = {}
-        msgcurrent = ''
-        with open(self.filename, 'r') as po_file:
+        checker = Checker()
+        with open(self.filename, 'r', encoding='utf-8') as po_file:
             for line in po_file:
-                numline += 1
-                line = line.strip()
-                if len(line) == 0:
-                    continue
-                if line.startswith('#,'):
-                    fuzzy = 'fuzzy' in line
-                    continue
-                if line.startswith('msg'):
-                    match = re.match(
-                        r'([a-zA-Z0-9-_]+(\[\d+\])?)[ \t](.*)',
-                        line)
-                    if match:
-                        oldmsgcurrent = msgcurrent
-                        msgcurrent = match.group(1)
-                        line = match.group(3)
-                        if msgcurrent == 'msgid':
-                            if oldmsgcurrent.startswith('msgstr'):
-                                self._add_message(numline_msgid,
-                                                  msgfuzzy,
-                                                  msg)
-                            msgfuzzy = fuzzy
-                            fuzzy = False
-                            msg = {}
-                            numline_msgid = numline
-                if msgcurrent and line.startswith('"'):
-                    msg[msgcurrent] = msg.get(msgcurrent, '') + line[1:-1]
-            if msgcurrent.startswith('msgstr'):
-                self._add_message(numline_msgid,
-                                  msgfuzzy,
-                                  msg)
+                message = checker.check_line(line.strip())
+                if message:
+                    self._add_message(*message)
+        message = checker.last_check()
+        if message:
+            self._add_message(*message)
 
     def compile(self):
         """
@@ -370,7 +461,7 @@ class PoFile(object):
         """
         output = ''
         try:
-            output = subprocess.check_output(
+            output = subprocess.check_output(  # nosec
                 ['msgfmt', '-c', '-o', '/dev/null', self.filename],
                 stderr=subprocess.STDOUT)
         except subprocess.CalledProcessError as exc:
@@ -378,7 +469,7 @@ class PoFile(object):
         return (output, 0)
 
 
-class PoCheck(object):
+class PoCheck:
     """Perform checks on a gettext file."""
 
     def __init__(self):
@@ -386,6 +477,7 @@ class PoCheck(object):
         self.checks = {
             'compile': True,
             'fuzzy': False,
+            'check_noqa': False,
             'lines': True,
             'punct': True,
             'whitespace': True,
@@ -400,25 +492,19 @@ class PoCheck(object):
         self.pwl = None
 
     def __repr__(self):
-        return ('checks: {self.checks}, '
-                'dicts: {self.dicts}, '
-                'extra_checkers: {self.extra_checkers}, '
-                'pwl: {self.pwl}'.format(self=self))
+        return (f'checks: {self.checks}, dicts: {self.dicts}, '
+                f'extra_checkers: {self.extra_checkers}')
 
     def set_check(self, check, state):
         """Enable/disable a specific check."""
         if check in self.checks:
             self.checks[check] = bool(state)
 
-    def set_spelling_options(self, spelling, dicts, pwl):
+    def set_spelling_options(self, spelling, dicts, pwl_files):
         """Set spelling options."""
         self.spelling = spelling
         self.dicts = dicts
-        self.pwl = pwl
-
-        # check if pwl file exists
-        if pwl and not os.path.isfile(pwl):
-            raise IOError('pwl file "{0}" not found'.format(pwl))
+        self.pwl = get_concatenated_files(pwl_files)
 
         # build extra checkers with dicts
         self.extra_checkers = []
@@ -431,8 +517,8 @@ class PoCheck(object):
                     _dict = Dict(lang)
                     self.extra_checkers.append(SpellChecker(_dict))
                 except DictNotFoundError:
-                    print('WARNING: enchant dictionary not found for '
-                          'language "{0}"'.format(lang))
+                    print(f'WARNING: enchant dictionary not found for '
+                          f'language "{lang}"')
 
     def _get_language_checker(self, po_file, reports):
         """Get checker for PO file language."""
@@ -444,14 +530,23 @@ class PoCheck(object):
             lang = po_file.props['language'] \
                 if self.spelling == 'str' else 'en'
             try:
-                _dict = DictWithPWL(lang, self.pwl)
+                if self.pwl:
+                    with tempfile.NamedTemporaryFile() as tmp_file:
+                        tmp_file.write(self.pwl.encode('utf-8'))
+                        tmp_file.flush()
+                        _dict = DictWithPWL(lang, tmp_file.name)
+                else:
+                    _dict = DictWithPWL(lang, None)
                 checker.append(SpellChecker(_dict))
             except DictNotFoundError:
-                reports.append(PoReport(
-                    'enchant dictionary not found for language "{0}"'
-                    ''.format(lang),
-                    'dict', po_file.filename,
-                    po_file.props['language_numline']))
+                reports.append(
+                    PoReport(
+                        f'enchant dictionary not found for language "{lang}"',
+                        'dict',
+                        po_file.filename,
+                        po_file.props['language_numline']
+                    )
+                )
                 checker = []
             except IOError as exc:
                 reports.append(PoReport(
@@ -460,6 +555,25 @@ class PoCheck(object):
                 checker = []
         return checker
 
+    def check_msg(self, po_file, checker, msg, reports):
+        """Check one message."""
+        if self.checks['extract']:
+            for mid, mstr in msg.messages:
+                if mid and mstr:
+                    reports.append(PoReport(mstr, 'extract'))
+        else:
+            if self.checks['lines']:
+                reports += msg.check_lines()
+            if self.checks['punct']:
+                reports += msg.check_punct(po_file.props['language'])
+            if self.checks['whitespace']:
+                reports += msg.check_whitespace()
+            if self.checks['whitespace_eol']:
+                reports += msg.check_whitespace_eol()
+            if self.spelling:
+                reports += msg.check_spelling(self.spelling, checker +
+                                              self.extra_checkers)
+
     def check_pofile(self, po_file):
         """
         Check translations in one PO file.
@@ -473,25 +587,13 @@ class PoCheck(object):
 
         # check all messages
         check_fuzzy = self.checks['fuzzy']
+        check_noqa = self.checks['check_noqa']
         for msg in po_file.msgs:
+            if msg.noqa and not check_noqa:
+                continue
             if msg.fuzzy and not check_fuzzy:
                 continue
-            if self.checks['extract']:
-                for mid, mstr in msg.messages:
-                    if mid and mstr:
-                        reports.append(PoReport(mstr, 'extract'))
-            else:
-                if self.checks['lines']:
-                    reports += msg.check_lines()
-                if self.checks['punct']:
-                    reports += msg.check_punct(po_file.props['language'])
-                if self.checks['whitespace']:
-                    reports += msg.check_whitespace()
-                if self.checks['whitespace_eol']:
-                    reports += msg.check_whitespace_eol()
-                if self.spelling:
-                    reports += msg.check_spelling(
-                        self.spelling, checker + self.extra_checkers)
+            self.check_msg(po_file, checker, msg, reports)
 
         return reports
 
@@ -514,7 +616,6 @@ class PoCheck(object):
                                          po_file.filename)]))
                 continue
             # compile the file (except if disabled)
-            compile_output = ''
             compile_rc = 0
             if self.checks['compile']:
                 compile_output, compile_rc = po_file.compile()
@@ -523,8 +624,7 @@ class PoCheck(object):
                 result.append((po_file.filename, self.check_pofile(po_file)))
             else:
                 # compilation failed
-                if sys.version_info >= (3,):
-                    compile_output = compile_output.decode('utf-8')
+                compile_output = bytes(compile_output).decode('utf-8')
                 result.append((po_file.filename,
                                [PoReport(compile_output, 'compile',
                                          po_file.filename)]))
